"""
Base directory for the package

Copyright 2021 InferStat Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Files and sub-directory information:

/algos - contains trading strategies
/data - sources of simulated and external data
/utilities - performance calculations and other utilities
__init__.py - package identifier
_version - package version number.  # TODO - refactor into package root __init__?
base.py

Created by: Joshua Mason
Created date: 11/03/2021
"""
